import dj_database_url

DATABASES = {
    'default': dj_database_url.config(default='sqlite:///db.sqlite'),
    'other': dj_database_url.config(env='DATABASE_URL_OTHER', default='sqlite:///db.sqlite'),
}

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.auth',
    'sql_server.pyodbc',
    'testapp',
)

SECRET_KEY = "django_tests_secret_key"

# Use a fast hasher to speed up tests.
PASSWORD_HASHERS = [
    'django.contrib.auth.hashers.MD5PasswordHasher',
]



from setuptools import find_packages, setup

CLASSIFIERS = [
    'License :: OSI Approved :: BSD License',
    'Framework :: Django',
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Framework :: Django :: 2.2',
    'Framework :: Django :: 3.0',
]

setup(
    name='django-mssql-backend-aad',
    version='0.1',
    description='Django backend for Microsoft SQL Server with AAD suport',
    long_description=open('README.rst').read(),
    author='Agustin Lucchetti',
    author_email='agustin.lucchetti@patagonian.it',
    url='https://bitbucket.org/patagoniantech/django-mssql-backend-aad',
    license='BSD',
    packages=find_packages(),
    install_requires=[
        'pyodbc>=3.0',
        'msal>=1.5.1'
    ],
    package_data={'sql_server.pyodbc': ['regex_clr.dll']},
    classifiers=CLASSIFIERS,
    keywords='django',
)



import struct
import msal


class AADAuthAccessTokenRetrievalError(Exception):
    pass


class AADAuth(object):
    """
    Defines the Azure Active Directory authentication class.
    """

    @classmethod
    def _get_access_token(cls, authority, client_id, scopes, secret):
        """
        Retrieves the application's access token through AAD.

        Parameters
        ----------
        authority : str
            The AAD authority url which is of the form
            `https://login.microsoftonline.com/<TENANT_ID>`.
        client_id : str
            The client identifier used to authenticate.
        scopest : list
            The list of scopes to use to authenticate. For database
            connections we should be using`["https://database.windows.net//.default"]`.
        secret : str
            The secret to authenticate with.

        Returns
        -------
        str
            The retrieved access token if present; otherwise, `None`.
        """
        app = msal.ConfidentialClientApplication(
            client_id, authority=authority, client_credential=secret
        )
        result = app.acquire_token_silent(scopes, account=None)
        if not result:
            result = app.acquire_token_for_client(scopes=scopes)
        access_token = result.get("access_token", None)
        return access_token

    @classmethod
    def _struct_from_access_token(cls, access_token):
        """
        Create a structure from the access token.

        Parameters
        ----------
        access_token : str
            The access token to convert.

        Returns
        -------
        str
            The expanded token structure required.
        """
        token_bytes = bytes(access_token, "UTF-8")
        expanded_token = b""
        for i in token_bytes:
            expanded_token += bytes({i})
            expanded_token += bytes(1)
        token_struct = struct.pack("=i", len(expanded_token)) + expanded_token
        return token_struct

    @classmethod
    def _construct_attrs_before_using_access_token(cls, access_token):
        """
        Constructs the `attrs_before` `pyodbc` parameter from an access token.

        Parameters
        ----------
        access_token : str
            The access token to construct the parameter with.

        Returns
        -------
        dict
            The `attrs_before` dictionary of values that will be used to authenticate
            using the access token.
        """
        SQL_COPT_SS_ACCESS_TOKEN = 1256
        token_struct = cls._struct_from_access_token(access_token)
        attrs_before = {SQL_COPT_SS_ACCESS_TOKEN: token_struct}
        return attrs_before

    @classmethod
    def create_attrs_before_with_access_token(cls, config):
        """
        Creates the `attrs_before` `pyodbc` parameter by first retrieving an AAD 
        access token for the application and then using it to build the parameter.

        Parameters
        ----------
        config : dict
            The dictionary of values needed to retrieve an application access token.
            Should contain `tenant_id`, `client_id` and `secret`.

        Returns
        -------
        dict
            The `attrs_before` dictionary of values that will be used to authenticate
            using the access token.

        Raises
        ------
        AADAuthAccessTokenRetrievalError
            When unable to retrieve a valid access token, the exception is raised.
        """
        tenant_id = config["tenant_id"]
        client_id = config["client_id"]
        secret = config["secret"]
        authority = "https://login.microsoftonline.com/" + tenant_id
        scopes = ["https://database.windows.net//.default"]
        access_token = cls._get_access_token(authority, client_id, scopes, secret)
        if not access_token:
            raise AADAuthAccessTokenRetrievalError("Unable to retrieve access token!")

        attrs_before = cls._construct_attrs_before_using_access_token(access_token)
        return attrs_before
